import django_tables2 as tables

from app.models import Book
from crud_views.lib.table import Table, UUIDLinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    ListViewTableMixin,
    DeleteViewPermissionRequired,
    ListViewPermissionRequired,
)
from crud_views.lib.viewset import ViewSet, ParentViewSet

cv_book = ViewSet(
    model=Book,
    name="book",
    parent=ParentViewSet(name="author"),
)


class BookTable(Table):
    id = UUIDLinkDetailColumn()
    title = tables.Column()
    price = tables.Column()
    author = tables.Column()


class BookListView(ListViewTableMixin, ListViewPermissionRequired):
    cv_viewset = cv_book
    # cv_list_actions = ["detail", "update", "delete"]

    table_class = BookTable


class BookDetailView(DetailViewPermissionRequired):
    cv_viewset = cv_book


class BookUpdateView(UpdateViewPermissionRequired):
    fields = ["title", "price"]
    cv_viewset = cv_book


class BookCreateView(CreateViewPermissionRequired):
    fields = ["title", "price"]
    cv_viewset = cv_book


class BookDeleteView(DeleteViewPermissionRequired):
    cv_viewset = cv_book



from typing import List, Type, Iterable

from django import forms
from django.http import HttpResponseRedirect
from django.views import generic
from polymorphic.models import PolymorphicModel

from crud_views.lib.check import CheckEitherAttribute, Check
from crud_views.lib.view import CrudView, CrudViewPermissionRequiredMixin
from crud_views.lib.settings import crud_views_settings
from .utils import get_polymorphic_child_models_content_types


class PolymorphicContentTypeForm(forms.Form):
    polymorphic_ctype_id = forms.ChoiceField(label="Type", choices=[])

    def __init__(self, *args, polymorphic_ctype_choices, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["polymorphic_ctype_id"].choices = polymorphic_ctype_choices


class PolymorphicCreateSelectView(CrudView, generic.FormView):
    template_name = "crud_views/view_create.html"
    form_class = PolymorphicContentTypeForm

    cv_key = "create_select"
    cv_object = False
    cv_path = "create/select"
    cv_success_key = "list"
    cv_context_actions = crud_views_settings.create_select_context_actions
    cv_polymorphic_exclude: List[Type[PolymorphicModel]] | None = None
    cv_polymorphic_include: List[Type[PolymorphicModel]] | None = None

    # texts and labels
    cv_header_template: str = "crud_views/snippets/header/create_select.html"
    cv_paragraph_template: str = "crud_views/snippets/paragraph/create_select.html"
    cv_action_label_template: str = "crud_views/snippets/action/create_select.html"
    cv_action_short_label_template: str = "crud_views/snippets/action_short/create_select.html"

    cv_icon_action = "fa-regular fa-square-plus"

    @classmethod
    def checks(cls) -> Iterable[Check]:
        yield from super().checks()
        yield CheckEitherAttribute(
            context=cls,
            id="E220",
            attribute1="cv_polymorphic_exclude",
            attribute2="cv_polymorphic_include",
            allow_none=True,
        )

    def get_form(self, form_class=None):
        form_class = form_class if form_class else self.get_form_class()
        content_types = get_polymorphic_child_models_content_types(self.model)  # noqa
        if self.cv_polymorphic_exclude:
            content_types = {k: v for k, v in content_types.items() if k not in self.cv_polymorphic_exclude}
        elif self.cv_polymorphic_include:
            content_types = {k: v for k, v in content_types.items() if k in self.cv_polymorphic_include}
        polymorphic_ctype_choices = [(ct.id, ct.name) for ct in content_types.values()]
        form_kwargs = self.get_form_kwargs()
        return form_class(polymorphic_ctype_choices=polymorphic_ctype_choices, **form_kwargs)

    def form_valid(self, form):
        polymorphic_ctype_id = form.cleaned_data["polymorphic_ctype_id"]
        kwargs = {"polymorphic_ctype_id": polymorphic_ctype_id}
        url = self.cv_get_url("create", extra_kwargs=kwargs)
        return HttpResponseRedirect(url)


class PolymorphicCreateSelectViewPermissionRequired(CrudViewPermissionRequiredMixin, PolymorphicCreateSelectView):
    cv_permission = "add"



from .context import ViewContext
from .base import CrudView, CrudViewPermissionRequiredMixin
from .buttons import ContextButton, ParentContextButton

__all__ = [
    "CrudView",
    "CrudViewPermissionRequiredMixin",
    "ViewContext",
    "ContextButton",
    "ParentContextButton",
]



import nox

DJANGO_VERSIONS = ["4.2", "5.2", "6.0"]


@nox.session(python=["3.12", "3.13", "3.14"], venv_backend="uv")
@nox.parametrize("django", DJANGO_VERSIONS)
def tests(session, django):
    # Django 4.2 does not support Python 3.14
    if django == "4.2" and session.python == "3.14":
        session.skip("Django 4.2 does not support Python 3.14")

    session.install(f"django~={django}.0")
    session.install(".[bootstrap5,polymorphic,workflow,test]")

    with session.chdir("./tests"):
        session.run("pytest", "-n", "auto", "--cov", "--cov-report=term-missing", *session.posargs)



from crud_views.lib.view import CrudViewPermissionRequiredMixin
from crud_views.lib.views import CreateView
from crud_views.lib.viewset import path_regs, PrimaryKeys
from .utils import PolymorphicCrudViewMixin


class PolymorphicCreateView(PolymorphicCrudViewMixin, CreateView):
    @classmethod
    def cv_path_contribute(cls) -> str:
        """
        Here we inject the polymorphic_ctype_id path.
        """
        path_contribute = path_regs.get_path_pk("polymorphic_ctype_id", PrimaryKeys.INT)
        return f"/ct/{path_contribute}/"


class PolymorphicCreateViewPermissionRequired(CrudViewPermissionRequiredMixin, PolymorphicCreateView):
    cv_permission = "add"



import django
import pytest
from django.conf import settings
from pathlib import Path

from django.test import Client

from tests.lib.helper.user import user_viewset_permission


def pytest_configure():
    settings.configure(
        BASE_DIR=Path(__file__).resolve().parent.parent,
        SECRET_KEY="django-testing",
        DEBUG=True,
        ALLOWED_HOSTS=[],
        INSTALLED_APPS=[
            "django.contrib.admin",
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.messages",
            "django.contrib.staticfiles",
            "django_bootstrap5",
            "crispy_forms",
            "crispy_bootstrap5",
            "polymorphic",
            "ordered_model",
            "django_tables2",
            "django_object_detail",
            "crud_views.apps.CrudViewsConfig",
            "django_fsm",
            "crud_views_workflow.apps.CrudViewsWorkflowConfig",
            "crud_views_polymorphic.apps.CrudViewsPolymorphicConfig",
            "tests.test1.app",
        ],
        MIDDLEWARE=[
            "django.middleware.security.SecurityMiddleware",
            "django.contrib.sessions.middleware.SessionMiddleware",
            "django.middleware.common.CommonMiddleware",
            "django.middleware.csrf.CsrfViewMiddleware",
            "django.contrib.auth.middleware.AuthenticationMiddleware",
            "django.contrib.messages.middleware.MessageMiddleware",
            "django.middleware.clickjacking.XFrameOptionsMiddleware",
        ],
        ROOT_URLCONF="tests.test1.project.urls",
        TEMPLATES=[
            {
                "BACKEND": "django.template.backends.django.DjangoTemplates",
                "DIRS": [],
                "APP_DIRS": True,
                "OPTIONS": {
                    "context_processors": [
                        "django.template.context_processors.debug",
                        "django.template.context_processors.request",
                        "django.contrib.auth.context_processors.auth",
                        "django.contrib.messages.context_processors.messages",
                        "crud_views.lib.context_processor.crud_views_context",
                    ],
                },
            },
        ],
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": ":memory:",
            }
        },
        # Internationalization
        LANGUAGE_CODE="en-us",
        TIME_ZONE="UTC",
        USE_I18N=True,
        USE_TZ=True,
        STATIC_URL="static/",
        DEFAULT_AUTO_FIELD="django.db.models.BigAutoField",
        # Django ViewSet configuration
        CRUD_VIEWS_EXTENDS="app/crud_views.html",
        CRUD_VIEWS_MANAGE_VIEWS_ENABLED="yes",
        # crispy
        CRISPY_TEMPLATE_PACK="bootstrap5",
        CRISPY_ALLOWED_TEMPLATE_PACKS="bootstrap5",
        # django_tables2
        DJANGO_TABLES2_TEMPLATE="django_tables2/bootstrap5.html",
    )

    django.setup()


@pytest.fixture
def user_a():
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_a", password="password")
    return user


@pytest.fixture
def cv_author():
    from tests.test1.app.views import cv_author as ret

    return ret


@pytest.fixture
def user_author_view(cv_author):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_author_view", password="password")

    user_viewset_permission(user, cv_author, "view")

    return user


@pytest.fixture
def client() -> Client:
    return Client()


@pytest.fixture
def client_user_author_view(client, user_author_view) -> Client:
    client.force_login(user_author_view)
    return client


@pytest.fixture
def user_author_change(cv_author):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_author_change", password="password")

    user_viewset_permission(user, cv_author, "change")

    return user


@pytest.fixture
def user_author_delete(cv_author):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_author_delete", password="password")

    user_viewset_permission(user, cv_author, "delete")

    return user


@pytest.fixture
def author_douglas_adams():
    from tests.test1.app.models import Author

    return Author.objects.create(first_name="Douglas", last_name="Adams")


@pytest.fixture
def user_author_add(cv_author):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_author_add", password="password")

    user_viewset_permission(user, cv_author, "add")

    return user


@pytest.fixture
def client_user_author_change(client, user_author_change) -> Client:
    client.force_login(user_author_change)
    return client


@pytest.fixture
def client_user_author_delete(client, user_author_delete) -> Client:
    client.force_login(user_author_delete)
    return client


@pytest.fixture
def client_user_author_add(client, user_author_add) -> Client:
    client.force_login(user_author_add)
    return client


@pytest.fixture
def client_user_a(client, user_a) -> Client:
    client.force_login(user_a)
    return client


@pytest.fixture
def author_terry_pratchett():
    from tests.test1.app.models import Author

    return Author.objects.create(first_name="Terry", last_name="Pratchett")


# --- Publisher fixtures ---


@pytest.fixture
def cv_publisher():
    from tests.test1.app.views import cv_publisher as ret

    return ret


@pytest.fixture
def publisher_penguin():
    from tests.test1.app.models import Publisher

    return Publisher.objects.create(name="Penguin")


@pytest.fixture
def publisher_harpercollins():
    from tests.test1.app.models import Publisher

    return Publisher.objects.create(name="HarperCollins")


@pytest.fixture
def user_publisher_view(cv_publisher):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_publisher_view", password="password")
    user_viewset_permission(user, cv_publisher, "view")
    return user


@pytest.fixture
def client_user_publisher_view(client, user_publisher_view) -> Client:
    client.force_login(user_publisher_view)
    return client


@pytest.fixture
def user_publisher_add(cv_publisher):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_publisher_add", password="password")
    user_viewset_permission(user, cv_publisher, "add")
    return user


@pytest.fixture
def client_user_publisher_add(client, user_publisher_add) -> Client:
    client.force_login(user_publisher_add)
    return client


@pytest.fixture
def user_publisher_change(cv_publisher):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_publisher_change", password="password")
    user_viewset_permission(user, cv_publisher, "change")
    return user


@pytest.fixture
def client_user_publisher_change(client, user_publisher_change) -> Client:
    client.force_login(user_publisher_change)
    return client


@pytest.fixture
def user_publisher_delete(cv_publisher):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_publisher_delete", password="password")
    user_viewset_permission(user, cv_publisher, "delete")
    return user


@pytest.fixture
def client_user_publisher_delete(client, user_publisher_delete) -> Client:
    client.force_login(user_publisher_delete)
    return client


# --- Book fixtures ---


@pytest.fixture
def cv_book():
    from tests.test1.app.views import cv_book as ret

    return ret


@pytest.fixture
def book_hitchhiker(publisher_penguin):
    from tests.test1.app.models import Book

    return Book.objects.create(title="The Hitchhiker's Guide to the Galaxy", publisher=publisher_penguin)


@pytest.fixture
def book_other_publisher(publisher_harpercollins):
    from tests.test1.app.models import Book

    return Book.objects.create(title="Other Book", publisher=publisher_harpercollins)


@pytest.fixture
def user_book_view(cv_book):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_book_view", password="password")
    user_viewset_permission(user, cv_book, "view")
    return user


@pytest.fixture
def client_user_book_view(client, user_book_view) -> Client:
    client.force_login(user_book_view)
    return client


@pytest.fixture
def user_book_add(cv_book):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_book_add", password="password")
    user_viewset_permission(user, cv_book, "add")
    return user


@pytest.fixture
def client_user_book_add(client, user_book_add) -> Client:
    client.force_login(user_book_add)
    return client


@pytest.fixture
def user_book_change(cv_book):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_book_change", password="password")
    user_viewset_permission(user, cv_book, "change")
    return user


@pytest.fixture
def client_user_book_change(client, user_book_change) -> Client:
    client.force_login(user_book_change)
    return client


@pytest.fixture
def user_book_delete(cv_book):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_book_delete", password="password")
    user_viewset_permission(user, cv_book, "delete")
    return user


@pytest.fixture
def client_user_book_delete(client, user_book_delete) -> Client:
    client.force_login(user_book_delete)
    return client


# --- Vehicle (polymorphic) fixtures ---


@pytest.fixture
def cv_vehicle():
    from tests.test1.app.views import cv_vehicle as ret

    return ret


@pytest.fixture
def car_sedan():
    from tests.test1.app.models import Car

    return Car.objects.create(name="Sedan", doors=4)


@pytest.fixture
def truck_semi():
    from tests.test1.app.models import Truck

    return Truck.objects.create(name="Semi", payload_tons=20)


@pytest.fixture
def user_vehicle_view(cv_vehicle):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_vehicle_view", password="password")
    user_viewset_permission(user, cv_vehicle, "view")
    return user


@pytest.fixture
def client_user_vehicle_view(client, user_vehicle_view) -> Client:
    client.force_login(user_vehicle_view)
    return client


@pytest.fixture
def user_vehicle_add(cv_vehicle):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_vehicle_add", password="password")
    user_viewset_permission(user, cv_vehicle, "add")
    return user


@pytest.fixture
def client_user_vehicle_add(client, user_vehicle_add) -> Client:
    client.force_login(user_vehicle_add)
    return client


@pytest.fixture
def user_vehicle_change(cv_vehicle):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_vehicle_change", password="password")
    user_viewset_permission(user, cv_vehicle, "change")
    return user


@pytest.fixture
def client_user_vehicle_change(client, user_vehicle_change) -> Client:
    client.force_login(user_vehicle_change)
    return client


@pytest.fixture
def user_vehicle_delete(cv_vehicle):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_vehicle_delete", password="password")
    user_viewset_permission(user, cv_vehicle, "delete")
    return user


@pytest.fixture
def client_user_vehicle_delete(client, user_vehicle_delete) -> Client:
    client.force_login(user_vehicle_delete)
    return client


# --- Campaign (workflow) fixtures ---


@pytest.fixture
def cv_campaign():
    from tests.test1.app.views import cv_campaign as ret

    return ret


@pytest.fixture
def campaign_new():
    from tests.test1.app.models import Campaign

    return Campaign.objects.create(name="Test Campaign")


@pytest.fixture
def campaign_active():
    from tests.test1.app.models import Campaign

    campaign = Campaign.objects.create(name="Active Campaign")
    campaign.wf_activate(by=None)
    campaign.save()
    return campaign


@pytest.fixture
def campaign_success():
    from tests.test1.app.models import Campaign

    campaign = Campaign.objects.create(name="Success Campaign")
    campaign.wf_activate(by=None)
    campaign.save()
    campaign.wf_done(by=None)
    campaign.save()
    return campaign


@pytest.fixture
def campaign_canceled():
    from tests.test1.app.models import Campaign

    campaign = Campaign.objects.create(name="Canceled Campaign")
    campaign.wf_cancel_new(by=None, comment="Canceled in test")
    campaign.save()
    return campaign


@pytest.fixture
def user_campaign_change(cv_campaign):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_campaign_change", password="password")
    user_viewset_permission(user, cv_campaign, "change")
    return user


@pytest.fixture
def client_user_campaign_change(client, user_campaign_change) -> Client:
    client.force_login(user_campaign_change)
    return client


@pytest.fixture
def user_campaign_view(cv_campaign):
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="user_campaign_view", password="password")
    user_viewset_permission(user, cv_campaign, "view")
    return user


@pytest.fixture
def client_user_campaign_view(client, user_campaign_view) -> Client:
    client.force_login(user_campaign_view)
    return client



from django.contrib.auth import get_user_model
from django.urls import reverse
from pydantic import BaseModel

from .context import ViewContext
from ..settings import crud_views_settings

User = get_user_model()


class ContextButton(BaseModel):
    """
    A context button is a button that is rendered in the context of a CrudView
    """

    key: str
    key_target: str | None = None
    label_template: str | None = None
    label_template_code: str | None = None

    def render_label(self, data: dict, context: ViewContext) -> str:
        if self.label_template:
            return context.view.render_snippet(data, self.label_template)
        elif self.label_template_code:
            return context.render_snippet(data, template_code=self.label_template_code)

    def get_context(self, context: ViewContext) -> dict:

        dict_kwargs = dict(cv_access=False, cv_url=context.view.cv_get_url(key=self.key_target, obj=context.object))

        # get target view class
        cls = context.view.cv_get_cls_assert_object(self.key_target, context.object)

        # check access
        if cls.cv_has_access(context.view.request.user, context.object):
            # get the url for the target key
            dict_kwargs.update(
                cv_access=True,
            )

        # final data
        data = cls.cv_get_dict(context=context, **dict_kwargs)

        # render action label
        cv_action_label = self.render_label(data, context)
        if cv_action_label:
            data["cv_action_label"] = cv_action_label

        return data


class ParentContextButton(ContextButton):
    """
    A context button that
    """

    def get_context(self, context: ViewContext) -> dict:

        # does the view has no parent?
        if not context.view.cv_viewset.parent:
            return dict()

        # get the parent view class, defined by target
        parent = context.view.cv_viewset.parent
        cls = parent.viewset.get_view_class(self.key_target)

        dict_kwargs = dict(cv_access=False, cv_icon_action=cls.cv_viewset.icon_header)

        # parent url kwargs
        kwargs = dict()
        for idx, arg in enumerate(context.view.cv_viewset.get_parent_url_args()):
            if idx == 0:
                if cls.cv_object:
                    kwargs[parent.viewset.pk_name] = context.view.kwargs[arg]
            else:
                kwargs[arg] = context.view.kwargs[arg]

        # parent url
        router_name = parent.viewset.get_router_name(self.key_target)
        cv_url = reverse(router_name, kwargs=kwargs)

        # get the url for the target key
        dict_kwargs.update(cv_url=cv_url)

        # check permission
        if cls.cv_has_access(context.view.request.user, context.object):
            dict_kwargs.update(
                cv_access=True,
            )

        data = cls.cv_get_dict(context=context, **dict_kwargs)
        return data


class FilterContextButton(ContextButton):
    """
    A context button that
    """

    key: str = "filter"

    def get_context(self, context: ViewContext) -> dict:
        from ..views import ListViewTableFilterMixin

        dict_kwargs = dict(cv_access=False)

        # if view has no filter, no button is shown
        if not isinstance(context.view, ListViewTableFilterMixin):
            return dict_kwargs

        # todo, check permission

        list_url = context.view.cv_get_url(key=context.view.cv_key)

        data = dict()

        # render action label
        cv_action_label = "Filter"  # todo, add render
        if cv_action_label:
            data["cv_action_label"] = cv_action_label

        data["cv_icon_action"] = "fa-solid fa-filter"
        # "fa-solid fa-filter-circle-xmark"
        # data["cv_url"] = "#filter-collapse"
        data["cv_url"] = list_url

        data["cv_template"] = f"{crud_views_settings.theme_path}/tags/context_action_filter.html"

        # data["cv_url"] = "javascript:alert(1);return false;"

        return data



# Generated by Django 4.2 on 2026-02-20 10:35

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="WorkflowInfo",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("transition", models.CharField(max_length=255)),
                ("state_old", models.CharField(max_length=255)),
                ("state_new", models.CharField(max_length=255)),
                ("comment", models.TextField(blank=True, null=True)),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                ("data", models.JSONField(blank=True, default=dict, null=True)),
                ("workflow_object_id", models.PositiveBigIntegerField()),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "workflow_object_content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
        ),
    ]



import django_tables2 as tables

import django_filters

from django.forms import modelform_factory

from tests.test1.app.models import Author, Publisher, Book, Vehicle, Car, Truck, Campaign
from crud_views.lib.crispy import CrispyModelViewMixin, CrispyDeleteForm, CrispyModelForm
from crud_views.lib.crispy.form import CrispyForm
from crud_views.lib.table import Table, UUIDLinkDetailColumn, LinkDetailColumn
from crud_views.lib.views import (
    ListViewTableMixin,
    ListViewTableFilterMixin,
    MessageMixin,
    ListViewPermissionRequired,
    DeleteViewPermissionRequired,
    CreateViewPermissionRequired,
    UpdateViewPermissionRequired,
    DetailViewPermissionRequired,
    CreateViewParentMixin,
    OrderedUpViewPermissionRequired,
    OrderedUpDownPermissionRequired,
)
from crud_views.lib.views.list import ListViewFilterFormHelper
from crud_views_polymorphic.lib import (
    PolymorphicCreateViewPermissionRequired,
    PolymorphicCreateSelectViewPermissionRequired,
    PolymorphicUpdateViewPermissionRequired,
    PolymorphicDetailViewPermissionRequired,
)
from crud_views_polymorphic.lib.create_select import PolymorphicContentTypeForm
from crud_views_polymorphic.lib.delete import PolymorphicDeleteViewPermissionRequired
from crud_views.lib.viewset import ViewSet, ParentViewSet
from crud_views_workflow.lib.forms import WorkflowForm
from crud_views_workflow.lib.views import WorkflowViewPermissionRequired
from crud_views.lib.crispy import Column4, Column6
from crispy_forms.layout import Row, Layout

cv_author = ViewSet(model=Author, name="author", icon_header="fa-regular fa-user")


class AuthorTable(Table):
    id = UUIDLinkDetailColumn()
    first_name = tables.Column()
    last_name = tables.Column()
    pseudonym = tables.Column()


class AuthorListView(ListViewTableMixin, ListViewPermissionRequired):
    table_class = AuthorTable

    cv_viewset = cv_author
    cv_list_actions = [
        "detail",
        "update",
        "delete",  # "up", "down", "redirect_child"
    ]


class AuthorDetailView(DetailViewPermissionRequired):
    cv_viewset = cv_author

    cv_property_display = [
        {
            "title": "Attributes",
            "properties": [
                "full_name",
                "first_name",
                "last_name",
                "pseudonym",
            ],
        },
    ]


class AuthorForm(CrispyModelForm):
    submit_label = "Create"

    class Meta:
        model = Author
        fields = ["first_name", "last_name", "pseudonym"]

    def get_layout_fields(self):
        return Row(Column4("first_name"), Column4("last_name"), Column4("pseudonym"))


class AuthorCreateView(CrispyModelViewMixin, CreateViewPermissionRequired):
    form_class = AuthorForm
    cv_viewset = cv_author


class AuthorUpdateView(CrispyModelViewMixin, UpdateViewPermissionRequired):
    form_class = AuthorForm
    cv_viewset = cv_author


class AuthorDeleteView(CrispyModelViewMixin, DeleteViewPermissionRequired):
    form_class = CrispyDeleteForm
    cv_viewset = cv_author


class AuthorUpView(OrderedUpViewPermissionRequired):
    cv_viewset = cv_author


class AuthorDownView(OrderedUpDownPermissionRequired):
    cv_viewset = cv_author


# --- Publisher (INT PK) ---

cv_publisher = ViewSet(
    model=Publisher,
    name="publisher",
)


class PublisherTable(Table):
    id = LinkDetailColumn()
    name = tables.Column()


class PublisherFilterFormHelper(ListViewFilterFormHelper):
    layout = Layout(Row(Column4("name")))


class PublisherFilter(django_filters.FilterSet):
    name = django_filters.CharFilter(lookup_expr="icontains")

    class Meta:
        model = Publisher
        fields = ["name"]


class PublisherListView(ListViewTableMixin, ListViewTableFilterMixin, ListViewPermissionRequired):
    table_class = PublisherTable
    filterset_class = PublisherFilter
    formhelper_class = PublisherFilterFormHelper
    cv_viewset = cv_publisher
    cv_list_actions = ["detail", "update", "delete"]


class PublisherDetailView(DetailViewPermissionRequired):
    cv_viewset = cv_publisher
    cv_property_display = [
        {
            "title": "Attributes",
            "properties": ["name"],
        },
    ]


class PublisherForm(CrispyModelForm):
    submit_label = "Create"

    class Meta:
        model = Publisher
        fields = ["name"]

    def get_layout_fields(self):
        return Row(Column4("name"))


class PublisherCreateView(CrispyModelViewMixin, MessageMixin, CreateViewPermissionRequired):
    form_class = PublisherForm
    cv_viewset = cv_publisher


class PublisherUpdateView(CrispyModelViewMixin, MessageMixin, UpdateViewPermissionRequired):
    form_class = PublisherForm
    cv_viewset = cv_publisher


class PublisherDeleteView(CrispyModelViewMixin, MessageMixin, DeleteViewPermissionRequired):
    form_class = CrispyDeleteForm
    cv_viewset = cv_publisher


# --- Book (INT PK, child of Publisher) ---

cv_book = ViewSet(
    model=Book,
    name="book",
    parent=ParentViewSet(name="publisher"),
)


class BookTable(Table):
    id = LinkDetailColumn()
    title = tables.Column()


class BookListView(ListViewTableMixin, ListViewPermissionRequired):
    table_class = BookTable
    cv_viewset = cv_book
    cv_list_actions = ["detail", "update", "delete"]


class BookDetailView(DetailViewPermissionRequired):
    cv_viewset = cv_book
    cv_property_display = [
        {
            "title": "Attributes",
            "properties": ["title"],
        },
    ]


class BookForm(CrispyModelForm):
    submit_label = "Create"

    class Meta:
        model = Book
        fields = ["title"]

    def get_layout_fields(self):
        return Row(Column4("title"))


class BookCreateView(CrispyModelViewMixin, CreateViewParentMixin, CreateViewPermissionRequired):
    form_class = BookForm
    cv_viewset = cv_book


class BookUpdateView(CrispyModelViewMixin, UpdateViewPermissionRequired):
    form_class = BookForm
    cv_viewset = cv_book


class BookDeleteView(CrispyModelViewMixin, DeleteViewPermissionRequired):
    form_class = CrispyDeleteForm
    cv_viewset = cv_book


# --- Vehicle (polymorphic) ---

cv_vehicle = ViewSet(
    model=Vehicle,
    name="vehicle",
)


class VehicleTable(Table):
    id = LinkDetailColumn()
    name = tables.Column()


class VehicleListView(ListViewTableMixin, ListViewPermissionRequired):
    table_class = VehicleTable
    cv_viewset = cv_vehicle
    cv_list_actions = ["detail", "update", "delete"]
    cv_context_actions = ["create_select"]


_CarForm = modelform_factory(Car, fields=["name", "doors"])


class CarForm(CrispyForm, _CarForm):
    def get_layout_fields(self):
        return Row(Column6("name"), Column6("doors"))


_TruckForm = modelform_factory(Truck, fields=["name", "payload_tons"])


class TruckForm(CrispyForm, _TruckForm):
    def get_layout_fields(self):
        return Row(Column6("name"), Column6("payload_tons"))


class CrispyVehicleContentTypeForm(CrispyForm, PolymorphicContentTypeForm):
    submit_label = "Select"

    def get_layout_fields(self):
        return Row(Column4("polymorphic_ctype_id"))


class VehicleCreateSelectView(CrispyModelViewMixin, PolymorphicCreateSelectViewPermissionRequired):
    form_class = CrispyVehicleContentTypeForm
    cv_viewset = cv_vehicle


class VehicleCreateView(CrispyModelViewMixin, PolymorphicCreateViewPermissionRequired):
    cv_viewset = cv_vehicle
    cv_context_actions = ["home"]
    polymorphic_forms = {
        Car: CarForm,
        Truck: TruckForm,
    }


class VehicleUpdateView(CrispyModelViewMixin, PolymorphicUpdateViewPermissionRequired):
    cv_viewset = cv_vehicle
    polymorphic_forms = {
        Car: CarForm,
        Truck: TruckForm,
    }


class VehicleDeleteView(CrispyModelViewMixin, PolymorphicDeleteViewPermissionRequired):
    form_class = CrispyDeleteForm
    cv_viewset = cv_vehicle


class VehicleDetailView(PolymorphicDetailViewPermissionRequired):
    cv_viewset = cv_vehicle
    cv_property_display = [
        {
            "title": "Attributes",
            "properties": ["name"],
        },
    ]


# --- Campaign (workflow) ---

cv_campaign = ViewSet(
    model=Campaign,
    name="campaign",
)


class CampaignTable(Table):
    id = LinkDetailColumn()
    name = tables.Column()
    state = tables.Column(accessor="state_badge")


class CampaignListView(ListViewTableMixin, ListViewPermissionRequired):
    cv_viewset = cv_campaign
    cv_list_actions = ["detail", "workflow", "update", "delete"]
    table_class = CampaignTable


class CampaignDetailView(DetailViewPermissionRequired):
    cv_viewset = cv_campaign
    cv_property_display = [
        {
            "title": "Attributes",
            "properties": ["name", "state"],
        },
    ]


class CampaignWorkflowForm(WorkflowForm):
    class Meta(WorkflowForm.Meta):
        model = Campaign


class CampaignWorkflowView(CrispyModelViewMixin, MessageMixin, WorkflowViewPermissionRequired):
    cv_context_actions = ["list", "detail", "workflow"]
    cv_viewset = cv_campaign
    form_class = CampaignWorkflowForm



# Generated by Django 5.1.4 on 2024-12-27 14:59

import django.db.models.deletion
import uuid
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("app", "0001_initial"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Poly",
            fields=[
                ("id", models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ("shared", models.CharField(max_length=100)),
                (
                    "polymorphic_ctype",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="polymorphic_%(app_label)s.%(class)s_set+",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
        ),
        migrations.CreateModel(
            name="PolyOne",
            fields=[
                (
                    "poly_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="app.poly",
                    ),
                ),
                ("one", models.CharField(max_length=100)),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
            bases=("app.poly",),
        ),
        migrations.CreateModel(
            name="PolyTwo",
            fields=[
                (
                    "poly_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="app.poly",
                    ),
                ),
                ("two", models.CharField(max_length=100)),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
            bases=("app.poly",),
        ),
    ]



from enum import IntEnum, StrEnum


class WorkflowComment(IntEnum):
    NONE = 0
    OPTIONAL = 1
    REQUIRED = 2


class BadgeEnum(StrEnum):
    """
    Bootstrap contextual colours for state badges.
    """

    PRIMARY = "primary"
    SECONDARY = "secondary"
    SUCCESS = "success"
    DANGER = "danger"
    WARNING = "warning"
    INFO = "info"
    LIGHT = "light"
    DARK = "dark"



from django.contrib.auth.views import LoginView, LogoutView
from django.urls import path

from app.views import IndexView
from app.views.author import cv_author
from app.views.book import cv_book
from app.views.bar import cv_bar
from app.views.baz import cv_baz
from app.views.foo import cv_foo
from app.views.poly import cv_poly

urlpatterns = [
    path("", IndexView.as_view(), name="index"),
    path("login/", LoginView.as_view(), name="login"),
    path("logout/", LogoutView.as_view(), name="logout"),
]

urlpatterns += (
    cv_author.urlpatterns
    + cv_book.urlpatterns
    + cv_baz.urlpatterns
    + cv_bar.urlpatterns
    + cv_foo.urlpatterns
    + cv_poly.urlpatterns
)



from crispy_forms.helper import FormHelper
from crispy_forms.layout import Row
from django.forms import modelform_factory

from app.models.poly import PolyTwo
from crud_views.lib.crispy import Column6
from crud_views.lib.crispy.form import CrispyForm

from crud_views.lib.formsets import Formsets

_PolyTwoForm = modelform_factory(
    PolyTwo,
    fields=[
        "shared",
        "two",
    ],
)


class PolyTwoForm(CrispyForm, _PolyTwoForm):
    def get_layout_fields(self):
        return [Row(Column6("shared"), Column6("two")), Row(Formsets())]

    @property
    def helper(self) -> FormHelper:
        # todo: is this needed?
        h = super().helper
        h.form_tag = False
        return h



from functools import wraps
from typing import Type


class ViewSetNotFoundError(Exception):
    pass


class ViewSetKeyFoundError(Exception):
    pass


class ViewSetError(Exception):
    pass


class CrudViewError(Exception):
    pass


class ParentViewSetError(Exception):
    pass


def cv_raise(expression: bool, msg: str, exception: Type[Exception] = ViewSetError):
    if not expression:
        raise exception(msg)


STRICT = False


def ignore_exception(exception_type, default_value=None, default_empty_dict: bool = False):
    """
    Ignore exception and return default value if strict is False
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except exception_type as e:
                # todo: get strict from settings
                if STRICT:
                    raise e
                if default_empty_dict:
                    return dict()
                return default_value

        return wrapper

    return decorator



from datetime import date

import django_tables2 as tables
from crispy_forms.layout import Row
from django.utils.translation import gettext as _

from app.models import Person
from crud_views.lib.crispy import Column4, CrispyModelForm, CrispyModelViewMixin, CrispyDeleteForm
from crud_views.lib.table import Table, LinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    MessageMixin,
    ListViewTableMixin,
    ListViewPermissionRequired,
    DeleteViewPermissionRequired,
    CreateViewParentMixin,
)
from crud_views.lib.viewset import ViewSet, ParentViewSet

cv_person = ViewSet(
    model=Person,
    name="members",
    icon_header="fa-regular fa-user",
    parent=ParentViewSet(
        name="group",
        many_to_many_through_attribute="members",
    ),
)


class PersonForm(CrispyModelForm):
    class Meta:
        model = Person
        fields = ["name"]

    def get_layout_fields(self):
        return Row(Column4("name"))


class PersonTable(Table):
    id = LinkDetailColumn(attrs=Table.ca.ID)
    name = tables.Column()
    date_joined = tables.Column(verbose_name=_("Date joined"), empty_values=())

    def render_date_joined(self, value, record):
        return record.membership_set.first().date_joined


class PersonListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Person

    cv_viewset = cv_person
    cv_list_actions = [
        "detail",
        "update",
        "delete",
        # "redirect_child"
    ]

    table_class = PersonTable


class PersonCreateView(CrispyModelViewMixin, MessageMixin, CreateViewParentMixin, CreateViewPermissionRequired):
    model = Person
    form_class = PersonForm
    cv_viewset = cv_person
    cv_message = "Created person »{object}«"

    def cv_parent_many_to_many_through_defaults(self, instance, parent_instance, m2m) -> dict:
        defaults = super().cv_parent_many_to_many_through_defaults(instance, parent_instance, m2m)
        defaults.update({"date_joined": date.today()})
        return defaults


class PersonUpdateView(CrispyModelViewMixin, MessageMixin, UpdateViewPermissionRequired):
    model = Person
    form_class = PersonForm
    cv_viewset = cv_person
    cv_message = "Updated person »{object}«"


class PersonDeleteView(CrispyModelViewMixin, MessageMixin, DeleteViewPermissionRequired):
    model = Person
    form_class = CrispyDeleteForm
    cv_viewset = cv_person
    cv_message = "Deleted person »{object}«"


class PersonDetailView(DetailViewPermissionRequired):
    model = Person
    cv_viewset = cv_person

    cv_property_display = [
        {
            "title": _("Attributes"),
            "icon": "user",
            "description": _("Member details"),
            "properties": [
                "id",
                "name",
            ],
        },
    ]



from django.contrib.auth import get_user_model

from crud_views.lib.exceptions import cv_raise

try:
    from django_filters.views import FilterMixin

    _base_metaclass = type(FilterMixin)
except ImportError:
    _base_metaclass = type

User = get_user_model()


class CrudViewMetaClass(_base_metaclass):
    """
    Registers CrudViews at ViewSet
    """

    renamed_attributes = ()

    def __new__(cls, name, bases, attrs, **kwargs):
        obj = super().__new__(cls, name, bases, attrs, **kwargs)
        cv_viewset = attrs.get("cv_viewset")
        if cv_viewset:
            # get key to register view
            key = getattr(obj, "cv_key", None)
            cv_raise(key is not None, f"ViewSet {obj} has no attribute cv_key")

            # register view
            cv_viewset.register_view_class(key, obj)  # noqa

            # auto-set model from viewset if not explicitly provided
            if "model" not in attrs:
                obj.model = cv_viewset.model
        return obj



"""
Tests for DetailView.cv_property_display and checks().
"""

from django_object_detail import PropertyConfig


def _errors(view_cls):
    return [msg for chk in view_cls.checks() for msg in chk.messages()]


def _error_ids(view_cls):
    return {e.id for e in _errors(view_cls)}


# ---------------------------------------------------------------------------
# cv_property_display property adapter
# ---------------------------------------------------------------------------


def test_property_display_returns_cv_property_display():
    """property_display property returns cv_property_display."""
    from crud_views.lib.views.detail import DetailView

    class TestView(DetailView):
        cv_key = "detail"
        cv_property_display = [{"title": "A", "properties": ["x"]}]

    view = TestView()
    assert view.property_display == [{"title": "A", "properties": ["x"]}]


def test_property_display_none_when_not_set():
    """property_display returns None when cv_property_display is not overridden."""
    from crud_views.lib.views.detail import DetailView

    class TestView(DetailView):
        cv_key = "detail"

    view = TestView()
    assert view.property_display is None


# ---------------------------------------------------------------------------
# checks(): happy path
# ---------------------------------------------------------------------------


def test_detail_view_checks_pass():
    """A correctly configured DetailView produces no check errors."""
    from tests.test1.app.views import AuthorDetailView

    assert _errors(AuthorDetailView) == []


def test_detail_view_checks_pass_with_property_config():
    """cv_property_display containing PropertyConfig objects passes checks."""
    from tests.test1.app.views import CampaignDetailView

    assert _errors(CampaignDetailView) == []


# ---------------------------------------------------------------------------
# checks(): E240 — cv_property_display missing / None
# ---------------------------------------------------------------------------


def test_detail_view_checks_missing_cv_property_display():
    """cv_property_display=None yields E240."""
    from crud_views.lib.views.detail import DetailView

    class NoDisplayView(DetailView):
        cv_key = "detail"
        cv_path = "detail"
        # cv_property_display deliberately not set

    assert "viewset.E240" in _error_ids(NoDisplayView)


# ---------------------------------------------------------------------------
# checks(): E241 — not a list
# ---------------------------------------------------------------------------


def test_detail_view_checks_not_a_list():
    """cv_property_display that is not a list yields E241."""
    from crud_views.lib.views.detail import DetailView

    class BadTypeView(DetailView):
        cv_key = "detail"
        cv_path = "detail"
        cv_property_display = "not a list"

    assert "viewset.E241" in _error_ids(BadTypeView)


# ---------------------------------------------------------------------------
# checks(): E242 — group missing 'title'
# ---------------------------------------------------------------------------


def test_detail_view_checks_group_missing_title():
    """A group dict without 'title' yields E242."""
    from crud_views.lib.views.detail import DetailView

    class MissingTitleView(DetailView):
        cv_key = "detail"
        cv_path = "detail"
        cv_property_display = [{"properties": ["x"]}]  # no 'title'

    assert "viewset.E242" in _error_ids(MissingTitleView)


# ---------------------------------------------------------------------------
# checks(): E243 — group missing 'properties'
# ---------------------------------------------------------------------------


def test_detail_view_checks_group_missing_properties():
    """A group dict without 'properties' yields E243."""
    from crud_views.lib.views.detail import DetailView

    class MissingPropsView(DetailView):
        cv_key = "detail"
        cv_path = "detail"
        cv_property_display = [{"title": "A"}]  # no 'properties'

    assert "viewset.E243" in _error_ids(MissingPropsView)


# ---------------------------------------------------------------------------
# checks(): E244 — 'properties' not a list
# ---------------------------------------------------------------------------


def test_detail_view_checks_properties_not_a_list():
    """'properties' that is not a list yields E244."""
    from crud_views.lib.views.detail import DetailView

    class BadPropsView(DetailView):
        cv_key = "detail"
        cv_path = "detail"
        cv_property_display = [{"title": "A", "properties": "not a list"}]

    assert "viewset.E244" in _error_ids(BadPropsView)


# ---------------------------------------------------------------------------
# checks(): E245 — invalid property type
# ---------------------------------------------------------------------------


def test_detail_view_checks_invalid_property_type():
    """A property that is not str, dict, or PropertyConfig yields E245."""
    from crud_views.lib.views.detail import DetailView

    class BadPropTypeView(DetailView):
        cv_key = "detail"
        cv_path = "detail"
        cv_property_display = [{"title": "A", "properties": [42]}]

    assert "viewset.E245" in _error_ids(BadPropTypeView)


# ---------------------------------------------------------------------------
# checks(): valid variants accepted
# ---------------------------------------------------------------------------


def test_detail_view_checks_string_property_accepted():
    """Plain string properties pass E245."""
    from crud_views.lib.views.detail import DetailView

    class StringPropView(DetailView):
        cv_key = "detail"
        cv_path = "detail"
        cv_property_display = [{"title": "A", "properties": ["name", "email"]}]

    assert "viewset.E245" not in _error_ids(StringPropView)


def test_detail_view_checks_dict_property_accepted():
    """Dict properties pass E245."""
    from crud_views.lib.views.detail import DetailView

    class DictPropView(DetailView):
        cv_key = "detail"
        cv_path = "detail"
        cv_property_display = [{"title": "A", "properties": [{"path": "name", "detail": "The name"}]}]

    assert "viewset.E245" not in _error_ids(DictPropView)


def test_detail_view_checks_property_config_accepted():
    """PropertyConfig instances pass E245."""
    from crud_views.lib.views.detail import DetailView

    class PropertyConfigView(DetailView):
        cv_key = "detail"
        cv_path = "detail"
        cv_property_display = [{"title": "A", "properties": [PropertyConfig(path="name")]}]

    assert "viewset.E245" not in _error_ids(PropertyConfigView)



from django import template
from django.contrib.auth import get_user_model

User = get_user_model()

register = template.Library()


@register.inclusion_tag(takes_context=True, filename="crud_views/formsets/formset.html")
def cv_x_formset(context, x_formset):
    data = {"x_formset": x_formset}
    return data



from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models

User = get_user_model()


class WorkflowInfo(models.Model):
    transition = models.CharField(max_length=255)
    state_old = models.CharField(max_length=255)
    state_new = models.CharField(max_length=255)
    comment = models.TextField(null=True, blank=True)
    user = models.ForeignKey(User, on_delete=models.SET_NULL, null=True)
    timestamp = models.DateTimeField(auto_now_add=True)
    data = models.JSONField(default=dict, blank=True, null=True)
    workflow_object_content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    workflow_object_pk = models.CharField(max_length=255)
    workflow_object = GenericForeignKey("workflow_object_content_type", "workflow_object_pk")

    class Meta:
        indexes = [
            models.Index(fields=["workflow_object_pk", "workflow_object_content_type"]),
        ]



# Generated by Django 5.2.11 on 2026-02-20 07:40

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("app", "0002_question_questionchoice_questionchoicetag_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="polyanswer",
            options={},
        ),
        migrations.AlterModelOptions(
            name="polyanswernumber",
            options={},
        ),
        migrations.AlterModelOptions(
            name="polyanswertext",
            options={},
        ),
        migrations.AlterModelOptions(
            name="polyone",
            options={},
        ),
        migrations.AlterModelOptions(
            name="polythree",
            options={},
        ),
        migrations.AlterModelOptions(
            name="polytwo",
            options={},
        ),
    ]



import django_filters
import django_tables2 as tables

from app.models import Author
from crud_views.lib.table import Table, LinkChildColumn, UUIDLinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    MessageMixin,
    ListViewTableMixin,
    ListViewTableFilterMixin,
    ListViewPermissionRequired,
    OrderedUpViewPermissionRequired,
    OrderedUpDownPermissionRequired,
    DeleteViewPermissionRequired,
)
from crud_views.lib.viewset import ViewSet

cv_author = ViewSet(
    model=Author,
    name="author",
)


class AuthorFilter(django_filters.FilterSet):
    first_name = django_filters.CharFilter(lookup_expr="icontains")
    last_name = django_filters.CharFilter(lookup_expr="icontains")

    class Meta:
        model = Author
        fields = [
            "first_name",
            "last_name",
        ]


class AuthorTable(Table):
    id = UUIDLinkDetailColumn(attrs=Table.col_attr.wID)
    first_name = tables.Column(attrs=Table.col_attr.w20)
    last_name = tables.Column(attrs=Table.col_attr.w30)
    pseudonym = tables.Column(attrs=Table.col_attr.w20)
    books = LinkChildColumn(name="book", verbose_name="Books", attrs=Table.col_attr.w10)


class AuthorListView(ListViewTableMixin, ListViewTableFilterMixin, ListViewPermissionRequired):
    filterset_class = AuthorFilter

    cv_viewset = cv_author
    cv_list_actions = ["detail", "update", "delete", "up", "down"]

    table_class = AuthorTable


class AuthorDetailView(DetailViewPermissionRequired):
    cv_viewset = cv_author
    cv_property_display = [
        {
            "title": "Attributes",
            "properties": [
                "full_name",
                "first_name",
                "last_name",
                "pseudonym",
                "book_count",
            ],
        },
    ]


class AuthorUpdateView(MessageMixin, UpdateViewPermissionRequired):
    fields = ["first_name", "last_name", "pseudonym"]
    cv_viewset = cv_author
    cv_message = "Updated author »{object}«"


class AuthorCreateView(MessageMixin, CreateViewPermissionRequired):
    fields = ["first_name", "last_name", "pseudonym"]
    cv_viewset = cv_author
    cv_message = "Created author »{object}«"


class AuthorDeleteView(MessageMixin, DeleteViewPermissionRequired):
    cv_viewset = cv_author
    cv_message = "Deleted author »{object}«"


class AuthorUpView(MessageMixin, OrderedUpViewPermissionRequired):
    cv_viewset = cv_author
    cv_message = "Successfully moved author »{object}« up"


class AuthorDownView(MessageMixin, OrderedUpDownPermissionRequired):
    cv_viewset = cv_author
    cv_message = "Successfully moved author »{object}« down"



from __future__ import annotations

import re
from collections import OrderedDict
from typing import Any, OrderedDict as OrderedDictType, List, Iterable, Tuple
from typing import Dict, Type, Self

from crud_views.lib.view import CrudView
from django.core.exceptions import ValidationError
from django.forms.forms import BaseForm
from django.forms.models import ModelForm, BaseInlineFormSet
from django.http.request import HttpRequest
from django.template.loader import render_to_string
from ordered_model.models import OrderedModel
from pydantic import BaseModel, Field, model_validator
from enum import Enum

from .x import XForm, XFormSet


class FormSet(BaseModel, arbitrary_types_allowed=True):
    class PK(str, Enum):
        INT = r"\d+"
        UUID = r"[0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}"

    key: str | None = None
    original_key: str | None = None
    title: str | None = None
    fields: List[str]  # todo: get from formset klass ?
    pk_field: str  # todo: get from formset klass ?
    cv_view: CrudView | None = None
    parent: Self | None = None
    klass: Type[BaseInlineFormSet]
    children: Dict[str, Self] = Field(default_factory=lambda: OrderedDict())
    path: str | None = None
    pk: PK = PK.INT

    def __str__(self):
        return f"FormSet({self.title})"

    __repr__ = __str__

    @model_validator(mode="after")
    def validate_formset(self) -> Self:

        from .inline_formset import BaseInlineFormSet

        if self.klass.can_order:
            if not issubclass(self.klass.model, OrderedModel):
                raise ValidationError(
                    f"FormSet '{self.key}' model not a subclass of OrderedModel but formset.can_order is True"
                )

        if not issubclass(self.klass, BaseInlineFormSet):
            raise ValidationError(f"FormSet '{self.key}' klass not a subclass of BaseInlineFormSet")

        return self

    @property
    def can_delete(self) -> bool:
        return self.klass.can_delete

    @property
    def can_delete_extra(self) -> bool:
        return self.klass.can_delete_extra

    @property
    def can_order(self) -> bool:
        return self.klass.can_order

    @property
    def edit_only(self) -> bool:
        return self.klass.edit_only

    @property
    def hierarchy(self) -> List[Self]:
        if self.parent is None:
            return [self.original_key]
        return self.parent.hierarchy + [self.original_key]

    def init(
        self,
        request: HttpRequest,
        forms: List[BaseForm],
        index: int = 0,
        parent: XForm | None = None,
        parent_prefix: str | None = None,
        level: int = 0,
    ) -> Iterable[XFormSet]:
        """
        Initialize the formset and all its children.
        This will be called recursively for all children.
        While iterating over the formset hierarchy it will create
            - a XFormSet instance for each formset
            - and a XForm instance for each form
        The X-Elements can be used to render the formsets in the template.
        """

        self.path = request.path

        # get parent prefix with -
        parent_prefix_ = f"{parent_prefix}-" if parent_prefix else ""

        for i, form in enumerate(forms):
            prefix_key = f"{parent_prefix_}{form.instance.pk}-{index}"
            prefix = f"{self.key}-{prefix_key}"

            kwargs = {
                "formset": self,
                "instance": form.instance,
                "prefix": prefix,
                "parent_form": form if level > 0 else None,  # todo
                # "form_kwargs": {"cv_view": None} # TODO HACK!!
            }
            if request.POST:
                kwargs["data"] = request.POST
                kwargs["files"] = request.FILES
            formset_instance = self.klass(**kwargs)

            x_formset = XFormSet(
                level=level,
                prefix_key=prefix_key,
                instance=formset_instance,
                formset=self,
                parent=parent,
                management_form=formset_instance.management_form,
            )

            for formset_instance_form_index, formset_instance_form in enumerate(formset_instance.forms):
                formset_instance_form_prefix_key = f"{prefix_key}-{formset_instance_form_index}"
                x_form = XForm(
                    level=level,
                    prefix_key=formset_instance_form_prefix_key,
                    form=formset_instance_form,
                    parent=x_formset,
                )
                x_formset.forms.append(x_form)

                for index, (key, child_formset) in enumerate(self.children.items()):
                    x_formsets = child_formset.init(
                        request=request,
                        forms=[x_form.form],
                        level=level + 1,
                        parent=x_form,
                        parent_prefix=formset_instance_form_prefix_key,
                        index=index,
                    )
                    x_form.formsets.extend(list(x_formsets))

            yield x_formset

    def template(
        self,
        pk: int | None,
        index: int,
        force_form_index: int | None = None,
        parent_prefix: str | None = None,
        level: int = 0,
        parent: XForm | None = None,
    ) -> Iterable[XFormSet]:
        """
        Similar to init, but this will be called to render the template.
        Yes, it's similar to init but not the same.
        """

        # get parent prefix with -
        parent_prefix_ = f"{parent_prefix}-" if parent_prefix else ""

        prefix_key = f"{parent_prefix_}{pk}-{index}"
        prefix = f"{self.key}-{prefix_key}"

        kwargs = {"formset": self, "prefix": prefix}
        formset_instance = self.klass(**kwargs)

        x_formset = XFormSet(
            level=level,
            prefix_key=prefix_key,
            instance=formset_instance,
            formset=self,
            parent=parent,
            management_form=formset_instance.management_form,
            start_at_rows=True if level == 0 else False,
        )

        # get the forms
        forms = formset_instance.forms
        if level == 0:
            assert force_form_index is not None, "force_form_index is required"
            # in first template level we only need the first form
            form = formset_instance.forms[0]
            # hack the form prefix, so it matches the insert position
            # this is fine, we do not need to use the pk format here,
            # because we only split away the form-index which is the last value
            form_prefix = form.prefix.rsplit("-", 1)[0] + f"-{force_form_index}"
            form.prefix = form_prefix
            forms = [form]
            # todo: limit number of formset_instance.forms to 1 ?

        for formset_instance_form_index, formset_instance_form in enumerate(forms):
            # force the form index on level 0
            form_index = force_form_index if level == 0 else formset_instance_form_index
            formset_instance_form_prefix_key = f"{prefix_key}-{form_index}"

            x_form = XForm(
                level=level,
                prefix_key=formset_instance_form_prefix_key,
                form=formset_instance_form,
                formset=self,
                # instance=formset_instance,
                parent=x_formset,
            )
            x_formset.forms.append(x_form)

            for index, (key, child_formset) in enumerate(self.children.items()):
                x_formsets = child_formset.template(
                    pk=None,  # all sub-forms are new, so they have no pk
                    index=index,
                    level=level + 1,
                    parent=x_form,
                    parent_prefix=formset_instance_form_prefix_key,
                )
                x_form.formsets.extend(list(x_formsets))

        yield x_formset

    def is_valid(self) -> Iterable[Tuple[Any, bool]]:
        for instance in self.instances:
            yield instance, instance.is_valid()
        for formset in self.children.values():
            yield from formset.is_valid()

    def save(self, commit: bool = True):
        for instance in self.instances:
            instance.save(commit=commit)
        for formset in self.children.values():
            formset.save(commit=commit)

    def patch(self, cv_view: CrudView):
        self.cv_view = cv_view
        for formset in self.children.values():
            formset.parent = self
            formset.patch(cv_view)


class FormSets(BaseModel, arbitrary_types_allowed=True):
    formsets: OrderedDictType[str, FormSet]
    cv_view: CrudView | None = None
    x_formsets: List[XFormSet] = Field(default_factory=lambda: list())
    js_data: dict = Field(default_factory=lambda: dict())
    scripts: bool = True

    @model_validator(mode="before")
    @classmethod
    def check_keys(cls, data: Any) -> Any:
        formsets = data.get("formsets", dict())
        for key in formsets.keys():
            if re.match(r"^[a-z\-]+$", key) is None:
                raise ValueError(f"Formset key must contain only lowercase letters and slashes, got {key}")
        return data

    @model_validator(mode="after")
    def apply_hierarchy(self) -> Self:
        # todo: check this
        def apply_key(formset: FormSet, key: str, original_key: str, parent: FormSet | None = None):
            formset.original_key = original_key
            if formset.key is None:
                formset.key = key
                if parent:
                    formset.parent = parent
            for child_key, child_formset in formset.children.items():
                apply_key(child_formset, key=f"{key}-{child_key}", original_key=child_key, parent=formset)

        for key, formset in self.formsets.items():
            apply_key(formset, key=key, original_key=key)

        return self

    # Expose dict-like methods for convenience
    def items(self):
        return self.formsets.items()

    def keys(self):
        return self.formsets.keys()

    def values(self):
        return self.formsets.values()

    def get(self, key, default=None):
        return self.formsets.get(key, default)

    def __getitem__(self, key):
        return self.formsets[key]

    def __iter__(self):
        return iter(self.formsets)

    def is_valid(self) -> Iterable[Tuple[Any, bool]]:
        for x_formset in self.x_formsets:
            yield from x_formset.is_valid()

    def save(self, commit: bool = True):
        for x_formset in self.x_formsets:
            x_formset.save(commit=commit)

    def init(self, request: HttpRequest, form: ModelForm, instance, with_template: bool = True):
        for key, formset in self.items():
            x_formsets = list(formset.init(request=request, forms=[form]))
            self.x_formsets.extend(x_formsets)

    def init_js_data(self, view):
        data = dict(
            path=view.request.path,
        )
        self.js_data = data

    def get_template(self, key_path: List[str], pk: int, num: int, parent_prefix=None):
        formset = None
        for level, key in enumerate(key_path):
            if level == 0:
                formset = self.get(key)
            else:
                formset = formset.children[key]

        x_formsets = list(formset.template(index=0, parent_prefix=parent_prefix, force_form_index=num, pk=pk))
        assert len(x_formsets) == 1
        x_formset = x_formsets[0]
        data = dict(x_formset=x_formset)

        rows = [form.prefix for form in x_formset.forms]
        html = render_to_string("crud_views/formsets/formset.html", data)
        data = dict(rows=rows, html=html)
        return data

    def clone(self, cv_view: CrudView) -> Self:
        formsets = self.model_copy(deep=True)
        formsets.cv_view = cv_view
        for key, formset in formsets.items():
            formset.patch(cv_view)
        return formsets
